@@ -560,7 +560,7 @@ basic_engine_shape (PangoEngineShape *engine,
 		    PangoFont        *font,
 		    const char       *text,
 		    gint              length,
-		    PangoAnalysis    *analysis,
+		    const PangoAnalysis *analysis,
 		    PangoGlyphString *glyphs)
 {
   int n_chars;
@@ -691,7 +691,7 @@ basic_engine_x_class_init (PangoEngineShapeClass *class)
 }
 
 PANGO_ENGINE_SHAPE_DEFINE_TYPE (BasicEngineX, basic_engine_x,
-				basic_engine_x_class_init, NULL);
+				basic_engine_x_class_init, NULL)
 
 void 
 PANGO_MODULE_ENTRY(init) (GTypeModule *module)
