@@ -360,7 +360,7 @@ pango_font_description_set_size (PangoFontDescription *desc,
  * @desc: a #PangoFontDescription
  * 
  * Gets the size field of a font description.
- * See pango_font_description_get_size().
+ * See pango_font_description_set_size().
  * 
  * Return value: the size field for the font description in points or device units.
  *   You must call pango_font_description_get_size_is_absolute()
@@ -802,7 +802,8 @@ find_field (const FieldMap *map, int n_elements, const char *str, int len, int *
   
   for (i=0; i<n_elements; i++)
     {
-      if (map[i].str && g_ascii_strncasecmp (map[i].str, str, len) == 0)
+      if (map[i].str[0] && g_ascii_strncasecmp (map[i].str, str, len) == 0 &&
+	  map[i].str[len] == '\0')
 	{
 	  if (val)
 	    *val = map[i].value;
@@ -847,16 +848,23 @@ getword (const char *str, const char *last, size_t *wordlen)
 static gboolean
 parse_size (const char *word,
 	    size_t      wordlen,
-	    int        *pango_size)
+	    int        *pango_size,
+	    gboolean   *size_is_absolute)
 {
   char *end;
   double size = g_ascii_strtod (word, &end);
 
-  if ((size_t)(end - word) == wordlen && size >= 0 && size <= 1000000) /* word is a valid float */
+  if (end != word &&
+      (end == word + wordlen ||
+       (end + 2 == word + wordlen && !strncmp (end, "px", 2))
+      ) && size >= 0 && size <= 1000000) /* word is a valid float */
     {
       if (pango_size)
 	*pango_size = (int)(size * PANGO_SCALE + 0.5);
 
+      if (size_is_absolute)
+        *size_is_absolute = end < word + wordlen;
+
       return TRUE;
     }
 
@@ -872,7 +880,8 @@ parse_size (const char *word,
  * comma separated list of families optionally terminated by a comma,
  * STYLE_OPTIONS is a whitespace separated list of words where each
  * WORD describes one of style, variant, weight, or stretch, and SIZE
- * is an decimal number (size in points). Any one of the options may
+ * is a decimal number (size in points) or optionally followed by the
+ * unit modifier "px" for absolute size. Any one of the options may
  * be absent.  If FAMILY-LIST is absent, then the family_name field of
  * the resulting font description will be initialized to %NULL.  If
  * STYLE-OPTIONS is missing, then all style options will be set to the
@@ -916,8 +925,10 @@ pango_font_description_from_string (const char *str)
    */
   if (wordlen != 0)
     {
-      if (parse_size (p, wordlen, &desc->size))
+      gboolean size_is_absolute;
+      if (parse_size (p, wordlen, &desc->size, &size_is_absolute))
 	{
+	  desc->size_is_absolute = size_is_absolute;
 	  desc->mask |= PANGO_FONT_MASK_SIZE;
 	  last = p;
 	}
@@ -1015,7 +1026,7 @@ pango_font_description_to_string (const PangoFontDescription  *desc)
       p = getword (desc->family_name, desc->family_name + strlen(desc->family_name), &wordlen);
       if (wordlen != 0 &&
 	  (find_field_any (p, wordlen, NULL) ||
-	   (parse_size (p, wordlen, NULL) &&
+	   (parse_size (p, wordlen, NULL, NULL) &&
 	    desc->weight == PANGO_WEIGHT_NORMAL &&
 	    desc->style == PANGO_STYLE_NORMAL &&
 	    desc->stretch == PANGO_STRETCH_NORMAL &&
@@ -1041,6 +1052,9 @@ pango_font_description_to_string (const PangoFontDescription  *desc)
 
       g_ascii_dtostr (buf, sizeof (buf), (double)desc->size / PANGO_SCALE);
       g_string_append (result, buf);
+
+      if (desc->size_is_absolute)
+	g_string_append (result, "px");
     }
   
   return g_string_free (result, FALSE);
@@ -1092,7 +1106,9 @@ pango_font_init (PangoFont *font)
  * pango_font_describe:
  * @font: a #PangoFont
  * 
- * Returns a description of the font.
+ * Returns a description of the font, with font size set in points.
+ * Use pango_font_describe_with_absolute_size() if you want the font
+ * size in device units.
  * 
  * Return value: a newly-allocated #PangoFontDescription object.
  **/
@@ -1104,6 +1120,32 @@ pango_font_describe (PangoFont      *font)
   return PANGO_FONT_GET_CLASS (font)->describe (font);
 }
 
+/**
+ * pango_font_describe_with_absolute_size:
+ * @font: a #PangoFont
+ * 
+ * Returns a description of the font, with absolute font size set
+ * (in device units). Use pango_font_describe() if you want the font
+ * size in points.
+ * 
+ * Return value: a newly-allocated #PangoFontDescription object.
+ *
+ * Since: 1.14
+ **/
+PangoFontDescription *
+pango_font_describe_with_absolute_size (PangoFont      *font)
+{
+  g_return_val_if_fail (font != NULL, NULL);
+
+  if (G_UNLIKELY (!PANGO_FONT_GET_CLASS (font)->describe_absolute))
+    {
+      g_warning ("describe_absolute not implemented for this font class, report this as a bug");
+      return pango_font_describe (font);
+    }
+
+  return PANGO_FONT_GET_CLASS (font)->describe_absolute (font);
+}
+
 /**
  * pango_font_get_coverage:
  * @font: a #PangoFont
@@ -1170,7 +1212,7 @@ pango_font_get_glyph_extents  (PangoFont      *font,
 			       PangoRectangle *ink_rect,
 			       PangoRectangle *logical_rect)
 {
-  if (G_UNLIKELY (!PANGO_IS_FONT (font)))
+  if (G_UNLIKELY (!font))
     {
       
       if (!_pango_warning_history.get_glyph_extents)
