@@ -30,7 +30,7 @@
 #include "pangocairo-private.h"
 #include "pangocairo-fc.h"
 #include "pangofc-private.h"
-#include "pango-utils.h"
+#include "pango-impl-utils.h"
 
 #define PANGO_TYPE_CAIRO_FC_FONT           (pango_cairo_fc_font_get_type ())
 #define PANGO_CAIRO_FC_FONT(object)        (G_TYPE_CHECK_INSTANCE_CAST ((object), PANGO_TYPE_CAIRO_FC_FONT, PangoCairoFcFont))
@@ -42,46 +42,11 @@ typedef struct _PangoCairoFcFont      PangoCairoFcFont;
 typedef struct _PangoCairoFcFontClass PangoCairoFcFontClass;
 
 typedef struct _GlyphExtentsCacheEntry    GlyphExtentsCacheEntry;
-typedef struct _GUnicharToGlyphCacheEntry GUnicharToGlyphCacheEntry;
-
-
-/* define one of these to profile one of the caches */
-#undef PROFILE_GLYPH_EXTENTS_CACHE
-#undef PROFILE_CHAR_TO_GLYPH_CACHE
-
-#undef PROFILE_GLYPH_CACHE
-#ifdef PROFILE_GLYPH_EXTENTS_CACHE
-# define PROFILE_GLYPH_CACHE GlyphExtentsCacheEntry
-#endif
-#ifdef PROFILE_CHAR_TO_GLYPH_CACHE
-# define PROFILE_GLYPH_CACHE GUnicharToGlyphCacheEntry
-#endif
-#ifdef PROFILE_GLYPH_CACHE
-static long num_cairo_fc_fonts;
-static long max_cairo_fc_fonts;
-static long num_glyph_caches;
-static long max_glyph_caches;
-static long num_glyph_cache_hits;
-static long num_glyph_cache_misses;
-#endif
-
-
 
 #define GLYPH_CACHE_NUM_ENTRIES 256 /* should be power of two */
 #define GLYPH_CACHE_MASK (GLYPH_CACHE_NUM_ENTRIES - 1)
 
-/* An entry in the fixed-size cache for the gunichar -> glyph mapping.
- * The cache is indexed by the lower N bits of the gunichar (see
- * GLYPH_CACHE_NUM_ENTRIES).  For scripts with few code points,
- * this should provide pretty much instant lookups.
- *
- * The "ch" field is zero if that cache entry is invalid.
- */
-struct _GUnicharToGlyphCacheEntry
-{
-  gunichar   ch;
-  PangoGlyph glyph;
-};
+#define PANGO_UNITS(Double) ((int)((Double) * PANGO_SCALE + 0.49999))
 
 /* An entry in the fixed-size cache for the glyph -> ink_rect mapping.
  * The cache is indexed by the lower N bits of the glyph (see
@@ -106,8 +71,7 @@ struct _PangoCairoFcFont
   cairo_font_options_t *options;
 
   PangoRectangle font_extents;
-  GlyphExtentsCacheEntry    *glyph_extents_cache;
-  GUnicharToGlyphCacheEntry *char_to_glyph_cache;
+  GlyphExtentsCacheEntry    *glyph_extents_cache;  
 };
 
 struct _PangoCairoFcFontClass
@@ -133,11 +97,11 @@ pango_cairo_fc_font_get_font_face (PangoCairoFont *font)
     {
       cffont->font_face = cairo_ft_font_face_create_for_pattern (fcfont->font_pattern);
       
-      /* Failure of the above should only occur for out of memory,
-       * we can't proceed at that point
+      /* Unable to create FT2 cairo scaled font.
+       * This means out of memory or a cairo/fontconfig/FreeType bug,
        */
       if (!cffont->font_face)
-	g_error ("Unable to create FT2 cairo font face.\nThis means out of memory or a cairo/fontconfig/FreeType bug");
+        return NULL;
     }
   
   return cffont->font_face;
@@ -154,16 +118,20 @@ pango_cairo_fc_font_get_scaled_font (PangoCairoFont *font)
 
       font_face = pango_cairo_fc_font_get_font_face (font);
 
+      if (!font_face)
+        return NULL;
+
       cffont->scaled_font = cairo_scaled_font_create (font_face,
 						      &cffont->font_matrix,
 						      &cffont->ctm,
 						      cffont->options);
 
-      /* Failure of the above should only occur for out of memory,
-       * we can't proceed at that point
+      /* Unable to create FT2 cairo scaled font.
+       * This means out of memory or a cairo/fontconfig/FreeType bug,
+       * or a missing font...
        */
       if (!cffont->scaled_font)
-	g_error ("Unable to create FT2 cairo scaled font.\nThis means out of memory or a cairo/fontconfig/FreeType bug");
+        return NULL;
     }
   
   return cffont->scaled_font;
@@ -196,7 +164,7 @@ cairo_font_iface_init (PangoCairoFontIface *iface)
 }
 
 G_DEFINE_TYPE_WITH_CODE (PangoCairoFcFont, pango_cairo_fc_font, PANGO_TYPE_FC_FONT,
-    { G_IMPLEMENT_INTERFACE (PANGO_TYPE_CAIRO_FONT, cairo_font_iface_init) });
+    { G_IMPLEMENT_INTERFACE (PANGO_TYPE_CAIRO_FONT, cairo_font_iface_init) })
 
 static void
 pango_cairo_fc_font_finalize (GObject *object)
@@ -214,27 +182,9 @@ pango_cairo_fc_font_finalize (GObject *object)
   if (cffont->glyph_extents_cache)
     {
       g_free (cffont->glyph_extents_cache);
-#ifdef PROFILE_GLYPH_EXTENTS_CACHE
-      num_glyph_caches--;
-      g_assert (num_glyph_caches >= 0);
-#endif
-    }
-
-  if (cffont->char_to_glyph_cache)
-    {
-      g_free (cffont->char_to_glyph_cache);
-#ifdef PROFILE_CHAR_TO_GLYPH_CACHE
-      num_glyph_caches--;
-      g_assert (num_glyph_caches >= 0);
-#endif
-    }
+    }  
 
   G_OBJECT_CLASS (pango_cairo_fc_font_parent_class)->finalize (object);
-
-#ifdef PROFILE_GLYPH_CACHE
-  num_cairo_fc_fonts--;
-  g_assert (num_cairo_fc_fonts >= 0);
-#endif
 }
 
 /* This function is cut-and-pasted from pangocairo-fcfont.c - it might be
@@ -292,6 +242,8 @@ pango_cairo_fc_font_lock_face (PangoFcFont *font)
 {
   PangoCairoFont *cfont = (PangoCairoFont *)font;
   cairo_scaled_font_t *scaled_font = pango_cairo_fc_font_get_scaled_font (cfont);
+  if (G_UNLIKELY (!scaled_font))
+    return NULL;
   
   return cairo_ft_scaled_font_lock_face (scaled_font);
 }
@@ -301,75 +253,29 @@ pango_cairo_fc_font_unlock_face (PangoFcFont *font)
 {
   PangoCairoFont *cfont = (PangoCairoFont *)font;
   cairo_scaled_font_t *scaled_font = pango_cairo_fc_font_get_scaled_font (cfont);
+  if (G_UNLIKELY (!scaled_font))
+    return;
   
   cairo_ft_scaled_font_unlock_face (scaled_font);
 }
 
-static PangoGlyph
-pango_cairo_fc_font_get_glyph (PangoFcFont *font,
-			       gunichar     wc)
-{
-  PangoCairoFcFont *cffont = (PangoCairoFcFont *)font;
-  guint idx;
-  GUnicharToGlyphCacheEntry *entry;
-
-  if (cffont->char_to_glyph_cache == NULL)
-    {
-      cffont->char_to_glyph_cache = g_new0 (GUnicharToGlyphCacheEntry, GLYPH_CACHE_NUM_ENTRIES);
-      /* Make sure all cache entries are invalid initially */
-      cffont->char_to_glyph_cache[0].ch = 1; /* char 1 cannot happen in bucket 0 */
-#ifdef PROFILE_CHAR_TO_GLYPH_CACHE
-      num_glyph_caches++;
-      if (num_glyph_caches > max_glyph_caches)
-	max_glyph_caches = num_glyph_caches;
-#endif
-    }
-
-  idx = wc & GLYPH_CACHE_MASK;
-  entry = cffont->char_to_glyph_cache + idx;
-
-  if (entry->ch != wc)
-    {
-#ifdef PROFILE_CHAR_TO_GLYPH_CACHE
-      num_glyph_cache_misses++;
-      /*g_message ("cache MISS: cffont %p, gunichar %x = '%c'", cffont, wc, (wc < 128) ? wc : 0);*/
-#endif
-      entry->ch = wc;
-      entry->glyph = ((PangoFcFontClass *) pango_cairo_fc_font_parent_class)->get_glyph (font, wc);
-    }
-#ifdef PROFILE_CHAR_TO_GLYPH_CACHE
-  else
-    {
-      num_glyph_cache_hits++;
-      /*g_message ("cache HIT: cffont %p, gunichar %x = '%c'", cffont, wc, (wc < 128) ? wc : 0);*/
-    }
-#endif
-
-  return entry->glyph;
-}
-
 static void
 pango_cairo_fc_font_glyph_extents_cache_init (PangoCairoFcFont *cffont)
 {
   PangoCairoFont *cfont = (PangoCairoFont *)cffont;
   cairo_scaled_font_t *scaled_font = pango_cairo_fc_font_get_scaled_font (cfont);
   cairo_font_extents_t font_extents;
+
   cairo_scaled_font_extents (scaled_font, &font_extents);
 
   cffont->font_extents.x = 0;
-  cffont->font_extents.y = - font_extents.ascent * PANGO_SCALE;
-  cffont->font_extents.height = (font_extents.ascent + font_extents.descent) * PANGO_SCALE;
+  cffont->font_extents.y = - PANGO_UNITS (font_extents.ascent);
+  cffont->font_extents.height = PANGO_UNITS (font_extents.ascent + font_extents.descent);
   cffont->font_extents.width = 0;
 
   cffont->glyph_extents_cache = g_new0 (GlyphExtentsCacheEntry, GLYPH_CACHE_NUM_ENTRIES);
   /* Make sure all cache entries are invalid initially */
   cffont->glyph_extents_cache[0].glyph = 1; /* glyph 1 cannot happen in bucket 0 */
-  
-#ifdef PROFILE_GLYPH_EXTENTS_CACHE
-  num_glyph_caches++;
-  if (num_glyph_caches > max_glyph_caches)
-    max_glyph_caches = num_glyph_caches;
-#endif
 }
 
 /* Fills in the glyph extents cache entry
@@ -390,11 +296,11 @@ compute_glyph_extents (PangoCairoFcFont       *cffont,
 				   &cairo_glyph, 1, &extents);
   
   entry->glyph = glyph;
-  entry->width = extents.x_advance * PANGO_SCALE;
-  entry->ink_rect.x = extents.x_bearing * PANGO_SCALE;
-  entry->ink_rect.y = extents.y_bearing * PANGO_SCALE;
-  entry->ink_rect.width = extents.width * PANGO_SCALE;
-  entry->ink_rect.height = extents.height * PANGO_SCALE;
+  entry->width = PANGO_UNITS (extents.x_advance);
+  entry->ink_rect.x = PANGO_UNITS (extents.x_bearing);
+  entry->ink_rect.y = PANGO_UNITS (extents.y_bearing);
+  entry->ink_rect.width = PANGO_UNITS (extents.width);
+  entry->ink_rect.height = PANGO_UNITS (extents.height);
 }
      
 static GlyphExtentsCacheEntry *
@@ -409,19 +315,8 @@ pango_cairo_fc_font_get_glyph_extents_cache_entry (PangoCairoFcFont       *cffon
 
   if (entry->glyph != glyph)
     {
-#ifdef PROFILE_GLYPH_EXTENTS_CACHE
-      num_glyph_cache_misses++;
-      /*g_message ("cache MISS: cffont %p, glyph %x", cffont, glyph);*/
-#endif
       compute_glyph_extents (cffont, glyph, entry);
     }
-#ifdef PROFILE_GLYPH_EXTENTS_CACHE
-  else
-    {
-      num_glyph_cache_hits++;
-      /*g_message ("cache HIT: cffont %p, glyph %x = '%c'", cffont, glyph);*/
-    }
-#endif
 
   return entry;
 }
@@ -435,6 +330,13 @@ pango_cairo_fc_font_get_glyph_extents (PangoFont      *font,
   PangoCairoFcFont *cffont = (PangoCairoFcFont *)font;
   GlyphExtentsCacheEntry *entry;
 
+  if (!pango_cairo_fc_font_get_scaled_font ((PangoCairoFont *)cffont))
+    {
+      /* Get generic unknown-glyph extents. */
+      pango_font_get_glyph_extents (NULL, glyph, ink_rect, logical_rect);
+      return;
+    }
+
   /* We need to initialize the cache here, since we use cffont->font_extents
    */
   if (cffont->glyph_extents_cache == NULL)
@@ -443,7 +345,7 @@ pango_cairo_fc_font_get_glyph_extents (PangoFont      *font,
   if (glyph == PANGO_GLYPH_EMPTY)
     {
       if (ink_rect)
-	*ink_rect = cffont->font_extents;
+	ink_rect->x = ink_rect->y = ink_rect->width = ink_rect->height = 0;
       if (logical_rect)
 	*logical_rect = cffont->font_extents;
       return;
@@ -481,19 +383,6 @@ pango_cairo_fc_font_shutdown (PangoFcFont *fcfont)
     }
 }
 
-#ifdef PROFILE_GLYPH_CACHE
-static void
-profile_glyph_cache_exit_cb (void)
-{
-  g_message ("Profiled %s cache", G_STRINGIFY (PROFILE_GLYPH_CACHE));
-  g_message ("Maximum number of PangoCairoFcFont objects: %ld", max_cairo_fc_fonts);
-  g_message ("Maximum number of glyph caches created: %ld (%ld KB)", max_glyph_caches, (max_glyph_caches * sizeof (PROFILE_GLYPH_CACHE) * GLYPH_CACHE_NUM_ENTRIES) / 1024);
-  g_message ("Glyph cache hits: %ld", num_glyph_cache_hits);
-  g_message ("Glyph cache misses: %ld", num_glyph_cache_misses);
-  g_message ("Glyph cache miss rate: %f%%", 100.0 * (double) num_glyph_cache_misses / (num_glyph_cache_misses + num_glyph_cache_hits));
-}
-#endif
-
 static void
 pango_cairo_fc_font_class_init (PangoCairoFcFontClass *class)
 {
@@ -508,22 +397,12 @@ pango_cairo_fc_font_class_init (PangoCairoFcFontClass *class)
 
   fc_font_class->lock_face = pango_cairo_fc_font_lock_face;
   fc_font_class->unlock_face = pango_cairo_fc_font_unlock_face;
-  fc_font_class->get_glyph = pango_cairo_fc_font_get_glyph;
   fc_font_class->shutdown = pango_cairo_fc_font_shutdown;
-
-#ifdef PROFILE_GLYPH_CACHE
-  atexit (profile_glyph_cache_exit_cb);
-#endif
 }
 
 static void
 pango_cairo_fc_font_init (PangoCairoFcFont *cffont)
 {
-#ifdef PROFILE_GLYPH_CACHE
-  num_cairo_fc_fonts++;
-  if (num_cairo_fc_fonts > max_cairo_fc_fonts)
-    max_cairo_fc_fonts = num_cairo_fc_fonts;
-#endif
 }
 
 /********************
@@ -622,5 +501,10 @@ _pango_cairo_fc_font_new (PangoCairoFcFontMap        *cffontmap,
 
   cffont->options = cairo_font_options_copy (_pango_cairo_context_get_merged_font_options (context));
   
+  /* fcfont's is_hinted controls metric hinting
+   */
+  PANGO_FC_FONT(cffont)->is_hinted = 
+    (cairo_font_options_get_hint_metrics(cffont->options) != CAIRO_HINT_METRICS_OFF);
+
   return PANGO_FC_FONT (cffont);
 }
