@@ -19,6 +19,8 @@
  * Boston, MA 02111-1307, USA.
  */
 
+#include <config.h>
+
 #include "pangocairo-private.h"
 
 typedef struct _PangoCairoRendererClass PangoCairoRendererClass;
@@ -55,6 +57,104 @@ set_color (PangoCairoRenderer *crenderer,
 			  color->green / 65535.,
 			  color->blue / 65535.);
 }
+
+static void
+_pango_cairo_renderer_draw_box_glyph (PangoCairoRenderer *crenderer,
+				      PangoGlyphInfo     *gi,
+				      double              cx,
+				      double              cy)
+{
+  double temp_x, temp_y;
+
+  cairo_save (crenderer->cr);
+  cairo_get_current_point (crenderer->cr, &temp_x, &temp_y);
+
+  cairo_rectangle (crenderer->cr,
+		   cx + 1.5,
+		   cy + 1.5 - PANGO_UNKNOWN_GLYPH_HEIGHT,
+		   (double)gi->geometry.width / PANGO_SCALE - 3.0,
+		   PANGO_UNKNOWN_GLYPH_HEIGHT - 3.0);
+
+  if (!crenderer->do_path)
+    {
+      cairo_set_line_width (crenderer->cr, 1.0);
+      cairo_stroke (crenderer->cr);
+    }
+
+  cairo_move_to (crenderer->cr, temp_x, temp_y);
+  cairo_restore (crenderer->cr);
+}
+
+static void
+_pango_cairo_renderer_draw_unknown_glyph (PangoCairoRenderer *crenderer,
+					  PangoFont          *font,
+					  PangoGlyphInfo     *gi,
+					  double              cx,
+					  double              cy)
+{
+  char buf[7];
+  double x0, y0;
+  int row, col;
+  int rows, cols;
+  char hexbox_string[2] = {0, 0};
+  double temp_x, temp_y;
+  PangoCairoHexBoxInfo *hbi;
+  gunichar ch;
+
+  cairo_save (crenderer->cr);
+  cairo_get_current_point (crenderer->cr, &temp_x, &temp_y);
+
+  hbi = _pango_cairo_font_get_hex_box_info ((PangoCairoFont *)font);      
+
+  if (!hbi || !_pango_cairo_font_install (PANGO_CAIRO_FONT (hbi->font), crenderer->cr))
+    {
+      _pango_cairo_renderer_draw_box_glyph (crenderer, gi, cx, cy);
+      goto done;
+    }
+
+  ch = gi->glyph & ~PANGO_GLYPH_UNKNOWN_FLAG;
+
+  rows = hbi->rows;
+  cols = (ch > 0xffff ? 6 : 4) / rows;
+  g_snprintf (buf, sizeof(buf), (ch > 0xffff) ? "%06X" : "%04X", ch);
+
+  cairo_rectangle (crenderer->cr,
+		   cx + hbi->pad_x * 1.5,
+		   cy + hbi->box_descent - hbi->box_height + hbi->pad_y * 0.5,
+		   (double)gi->geometry.width / PANGO_SCALE - 3 * hbi->pad_x,
+		   (hbi->box_height - hbi->pad_y));
+
+  if (!crenderer->do_path)
+    {
+      cairo_set_line_width (crenderer->cr, hbi->line_width);
+      cairo_stroke (crenderer->cr);
+    }
+
+  x0 = cx + hbi->pad_x * 3.0;
+  y0 = cy + hbi->box_descent - hbi->pad_y * 2;
+
+  for (row = 0; row < rows; row++)
+    {
+      double y = y0 - (rows - 1 - row) * (hbi->digit_height + hbi->pad_y);
+      for (col = 0; col < cols; col++)
+	{
+	  double x = x0 + col * (hbi->digit_width + hbi->pad_x);
+
+	  cairo_move_to (crenderer->cr, x, y);
+
+	  hexbox_string[0] = buf[row * cols + col];
+
+	  if (crenderer->do_path)
+	      cairo_text_path (crenderer->cr, hexbox_string);
+	  else
+	      cairo_show_text (crenderer->cr, hexbox_string);
+	}
+    }
+
+done:
+  cairo_move_to (crenderer->cr, temp_x, temp_y);
+  cairo_restore (crenderer->cr);
+}
      
 static void
 pango_cairo_renderer_draw_glyphs (PangoRenderer     *renderer,
@@ -73,11 +173,29 @@ pango_cairo_renderer_draw_glyphs (PangoRenderer     *renderer,
   cairo_glyph_t *cairo_glyphs;
   cairo_glyph_t stack_glyphs[MAX_STACK];
 
+  cairo_save (crenderer->cr);
   if (!crenderer->do_path)
-    {
-      cairo_save (crenderer->cr);
+    set_color (crenderer, PANGO_RENDER_PART_FOREGROUND);
 
-      set_color (crenderer, PANGO_RENDER_PART_FOREGROUND);
+  if (!_pango_cairo_font_install (PANGO_CAIRO_FONT (font), crenderer->cr))
+    {
+      for (i = 0; i < glyphs->num_glyphs; i++)
+	{
+	  PangoGlyphInfo *gi = &glyphs->glyphs[i];
+
+	  if (gi->glyph != PANGO_GLYPH_EMPTY)
+	    {
+	      double cx = crenderer->x_offset +
+			  (double)(x + x_position + gi->geometry.x_offset) / PANGO_SCALE;
+	      double cy = crenderer->y_offset +
+			  (double)(y + gi->geometry.y_offset) / PANGO_SCALE;
+
+	      _pango_cairo_renderer_draw_unknown_glyph (crenderer, font, gi, cx, cy);
+	    }	  
+	  x_position += gi->geometry.width;
+	}
+
+      goto done;
     }
 
   if (glyphs->num_glyphs > MAX_STACK)
@@ -90,20 +208,26 @@ pango_cairo_renderer_draw_glyphs (PangoRenderer     *renderer,
     {
       PangoGlyphInfo *gi = &glyphs->glyphs[i];
 
-      if (gi->glyph)
-	{
-	  cairo_glyphs[count].index = gi->glyph;
-	  cairo_glyphs[count].x = crenderer->x_offset + (double)(x + x_position + gi->geometry.x_offset) / PANGO_SCALE;
-	  cairo_glyphs[count].y = crenderer->y_offset + (double)(y + gi->geometry.y_offset) / PANGO_SCALE;
+      if (gi->glyph != PANGO_GLYPH_EMPTY)
+        {
+          double cx = crenderer->x_offset +
+		      (double)(x + x_position + gi->geometry.x_offset) / PANGO_SCALE;
+          double cy = crenderer->y_offset +
+		      (double)(y + gi->geometry.y_offset) / PANGO_SCALE;
 
-	  count++;
-	}
-	  
+          if (gi->glyph & PANGO_GLYPH_UNKNOWN_FLAG)
+	    _pango_cairo_renderer_draw_unknown_glyph (crenderer, font, gi, cx, cy);
+          else
+            {
+              cairo_glyphs[count].index = gi->glyph;
+              cairo_glyphs[count].x = cx;
+              cairo_glyphs[count].y = cy;
+              count++;
+            }
+        }	  
       x_position += gi->geometry.width;
     }
 
-  _pango_cairo_font_install (PANGO_CAIRO_FONT (font), crenderer->cr);
-  
   if (crenderer->do_path)
     cairo_glyph_path (crenderer->cr, cairo_glyphs, count);
   else
@@ -112,8 +236,8 @@ pango_cairo_renderer_draw_glyphs (PangoRenderer     *renderer,
   if (glyphs->num_glyphs > MAX_STACK)
     g_free (cairo_glyphs);
 
-  if (!crenderer->do_path)
-    cairo_restore (crenderer->cr);
+done:
+  cairo_restore (crenderer->cr);
   
 #undef MAX_STACK
 }
@@ -139,11 +263,13 @@ pango_cairo_renderer_draw_rectangle (PangoRenderer     *renderer,
 		   crenderer->x_offset + (double)x / PANGO_SCALE,
 		   crenderer->y_offset + (double)y / PANGO_SCALE, 
 		   (double)width / PANGO_SCALE, (double)height / PANGO_SCALE);
+
   if (!crenderer->do_path)
-    cairo_fill (crenderer->cr);
-  
-  if (!crenderer->do_path)
-    cairo_restore (crenderer->cr);
+    {
+      cairo_fill (crenderer->cr);
+    
+      cairo_restore (crenderer->cr);
+    }
 }
 
 /* Draws an error underline that looks like one of:
@@ -157,8 +283,8 @@ pango_cairo_renderer_draw_rectangle (PangoRenderer     *renderer,
  *        \  /    \  /              \  /    \  /          |
  *         \/      \/                \/      \/           -
  *         B                         B       
- * |----|
- *   unit_width = (HEIGHT_SQUARES - 1) * square
+ *         |---|
+ *       unit_width = (HEIGHT_SQUARES - 1) * square
  *
  * The x, y, width, height passed in give the desired bounding box;
  * x/width are adjusted to make the underline a integer number of units
@@ -166,7 +292,6 @@ pango_cairo_renderer_draw_rectangle (PangoRenderer     *renderer,
  */
 #define HEIGHT_SQUARES 2.5
 
-/* This code is cut-and-pasted between here and gtk+/gdk/gdkpango.c */
 static void
 draw_error_underline (cairo_t *cr,
 		      double   x,
@@ -180,7 +305,7 @@ draw_error_underline (cairo_t *cr,
   double y_top, y_bottom;
   int i;
 
-  x += (width - width_units * unit_width);
+  x += (width - width_units * unit_width) / 2;
   width = width_units * unit_width;
 
   y_top = y;
@@ -269,6 +394,146 @@ pango_cairo_renderer_class_init (PangoCairoRendererClass *klass)
   renderer_class->draw_error_underline = pango_cairo_renderer_draw_error_underline;
 }
 
+static PangoCairoRenderer *cached_renderer = NULL;
+static GStaticMutex cached_renderer_mutex = G_STATIC_MUTEX_INIT;
+
+static PangoCairoRenderer *
+acquire_renderer (void)
+{
+  PangoCairoRenderer *renderer;
+
+  /* renderer = _pango_cairo_font_map_get_renderer (PANGO_CAIRO_FONT_MAP (fontmap)); */
+
+  if (G_LIKELY (g_static_mutex_trylock (&cached_renderer_mutex)))
+    {
+      if (G_UNLIKELY (!cached_renderer))
+        cached_renderer = g_object_new (PANGO_TYPE_CAIRO_RENDERER, NULL);
+
+      renderer = cached_renderer;
+    }
+  else
+    renderer = g_object_new (PANGO_TYPE_CAIRO_RENDERER, NULL);
+
+  return renderer;
+}
+
+release_renderer (PangoCairoRenderer *renderer)
+{
+  if (G_LIKELY (renderer == cached_renderer))
+    {
+      renderer->cr = NULL;
+      renderer->do_path = FALSE;
+      renderer->x_offset = 0.;
+      renderer->y_offset = 0.;
+
+      g_static_mutex_unlock (&cached_renderer_mutex);
+    }
+  else
+    g_object_unref (renderer);
+}
+
+
+/* convenience wrappers using the default renderer */
+
+
+static void
+_pango_cairo_do_glyph_string (cairo_t          *cr,
+			      PangoFont        *font,
+			      PangoGlyphString *glyphs,
+			      gboolean          do_path)
+{
+  PangoCairoRenderer *crenderer = acquire_renderer ();
+  PangoRenderer *renderer = (PangoRenderer *) crenderer;
+
+  crenderer->cr = cr;
+  crenderer->do_path = do_path;
+  cairo_get_current_point (cr, &crenderer->x_offset, &crenderer->y_offset);
+
+  if (!do_path)
+    {
+      /* unset all part colors, since when drawing just a glyph string,
+       * prepare_run() isn't called.
+       */
+
+      pango_renderer_activate (renderer); 	 
+	   	 
+      pango_renderer_set_color (renderer, PANGO_RENDER_PART_FOREGROUND, NULL); 	 
+      pango_renderer_set_color (renderer, PANGO_RENDER_PART_BACKGROUND, NULL); 	 
+      pango_renderer_set_color (renderer, PANGO_RENDER_PART_UNDERLINE, NULL); 	 
+      pango_renderer_set_color (renderer, PANGO_RENDER_PART_STRIKETHROUGH, NULL); 	 
+    }
+  
+  pango_renderer_draw_glyphs (renderer, font, glyphs, 0, 0);
+
+  if (!do_path)
+    {
+      pango_renderer_deactivate (renderer);
+    }
+  
+  release_renderer (crenderer);
+}
+
+static void
+_pango_cairo_do_layout_line (cairo_t          *cr,
+			     PangoLayoutLine  *line,
+			     gboolean          do_path)
+{
+  PangoCairoRenderer *crenderer = acquire_renderer ();
+  PangoRenderer *renderer = (PangoRenderer *) crenderer;
+
+  crenderer->cr = cr;
+  crenderer->do_path = do_path;
+  cairo_get_current_point (cr, &crenderer->x_offset, &crenderer->y_offset);
+
+  pango_renderer_draw_layout_line (renderer, line, 0, 0);
+  
+  release_renderer (crenderer);
+}
+
+static void
+_pango_cairo_do_layout (cairo_t     *cr,
+			PangoLayout *layout,
+			gboolean     do_path)
+{
+  PangoCairoRenderer *crenderer = acquire_renderer ();
+  PangoRenderer *renderer = (PangoRenderer *) crenderer;
+  
+  crenderer->cr = cr;
+  crenderer->do_path = do_path;
+  cairo_get_current_point (cr, &crenderer->x_offset, &crenderer->y_offset);
+
+  pango_renderer_draw_layout (renderer, layout, 0, 0);
+  
+  release_renderer (crenderer);
+}
+
+static void 
+_pango_cairo_do_error_underline (cairo_t *cr,                                  
+				 double   x,
+				 double   y,
+				 double   width,
+				 double   height,
+				 gboolean do_path)
+{  
+  /* We don't use a renderer here, for a simple reason:
+   * the only renderer we can get is the default renderer, that
+   * is all implemented here, so we shortcircuit and make our
+   * life way easier.
+   */
+
+  if (!do_path)
+    cairo_new_path (cr);
+
+  draw_error_underline (cr, x, y, width, height);
+
+  if (!do_path)
+    cairo_fill (cr);
+} 
+
+
+/* public wrapper of above to show or append path */
+
+
 /**
  * pango_cairo_show_glyph_string:
  * @cr: a Cairo context
@@ -286,31 +551,10 @@ pango_cairo_show_glyph_string (cairo_t          *cr,
 			       PangoFont        *font,
 			       PangoGlyphString *glyphs)
 {
-  PangoFontMap *fontmap;
-  PangoCairoRenderer *crenderer;
-  PangoRenderer *renderer;
-
   g_return_if_fail (cr != NULL);
-  g_return_if_fail (PANGO_IS_CAIRO_FONT (font));
   g_return_if_fail (glyphs != NULL);
 
-  fontmap = pango_font_get_font_map (font);
-  renderer = _pango_cairo_font_map_get_renderer (PANGO_CAIRO_FONT_MAP (fontmap));
-  crenderer = PANGO_CAIRO_RENDERER (renderer);
-
-  cairo_save (cr);
-
-  crenderer->cr = cr;
-  crenderer->do_path = FALSE;
-  cairo_get_current_point (cr, &crenderer->x_offset, &crenderer->y_offset);
-  
-  pango_renderer_draw_glyphs (renderer, font, glyphs, 0, 0);
-  
-  crenderer->cr = NULL;
-  crenderer->x_offset = 0.;
-  crenderer->y_offset = 0.;
-  
-  cairo_restore (cr);
+  _pango_cairo_do_glyph_string (cr, font, glyphs, FALSE);
 }
 
 /**
@@ -328,32 +572,10 @@ void
 pango_cairo_show_layout_line (cairo_t          *cr,
 			      PangoLayoutLine  *line)
 {
-  PangoContext *context;
-  PangoFontMap *fontmap;
-  PangoRenderer *renderer;
-  PangoCairoRenderer *crenderer;
-
   g_return_if_fail (cr != NULL);
   g_return_if_fail (line != NULL);
 
-  context = pango_layout_get_context (line->layout);
-  fontmap = pango_context_get_font_map (context);
-  renderer = _pango_cairo_font_map_get_renderer (PANGO_CAIRO_FONT_MAP (fontmap));
-  crenderer = PANGO_CAIRO_RENDERER (renderer);
-
-  cairo_save (cr);
-
-  crenderer->cr = cr;
-  crenderer->do_path = FALSE;
-  cairo_get_current_point (cr, &crenderer->x_offset, &crenderer->y_offset);
-
-  pango_renderer_draw_layout_line (renderer, line, 0, 0);
-  
-  crenderer->cr = NULL;
-  crenderer->x_offset = 0.;
-  crenderer->y_offset = 0.;
-  
-  cairo_restore (cr);
+  _pango_cairo_do_layout_line (cr, line, FALSE);
 }
 
 /**
@@ -371,34 +593,41 @@ void
 pango_cairo_show_layout (cairo_t     *cr,
 			 PangoLayout *layout)
 {
-  PangoContext *context;
-  PangoFontMap *fontmap;
-  PangoRenderer *renderer;
-  PangoCairoRenderer *crenderer;
-
   g_return_if_fail (cr != NULL);
   g_return_if_fail (PANGO_IS_LAYOUT (layout));
 
-  context = pango_layout_get_context (layout);
-  fontmap = pango_context_get_font_map (context);
-  renderer = _pango_cairo_font_map_get_renderer (PANGO_CAIRO_FONT_MAP (fontmap));
-  crenderer = PANGO_CAIRO_RENDERER (renderer);
-  
-  cairo_save (cr);
-
-  crenderer->cr = cr;
-  crenderer->do_path = FALSE;
-  cairo_get_current_point (cr, &crenderer->x_offset, &crenderer->y_offset);
-
-  pango_renderer_draw_layout (renderer, layout, 0, 0);
-  
-  crenderer->cr = NULL;
-  crenderer->x_offset = 0.;
-  crenderer->y_offset = 0.;
-  
-  cairo_restore (cr);
+  _pango_cairo_do_layout (cr, layout, FALSE);
 }
 
+/**
+ * pango_cairo_show_error_underline:
+ * @cr: a Cairo context
+ * @x: The X coordinate of one corner of the rectangle
+ * @y: The Y coordinate of one corner of the rectangle
+ * @width: Non-negative width of the rectangle
+ * @height: Non-negative height of the rectangle
+ * 
+ * Draw a squiggly line in the specified cairo context that approximately
+ * covers the given rectangle in the style of an underline used to indicate a
+ * spelling error.  (The width of the underline is rounded to an integer
+ * number of up/down segments and the resulting rectangle is centered in the
+ * original rectangle)
+ *
+ * Since: 1.14
+ **/
+void 
+pango_cairo_show_error_underline (cairo_t *cr,                                  
+			          double  x,
+			          double  y,
+			          double  width,
+			          double  height)
+{  
+  g_return_if_fail (cr != NULL);
+  g_return_if_fail ((width >= 0) && (height >= 0));  
+  
+  _pango_cairo_do_error_underline (cr, x, y, width, height, FALSE);                                 
+} 
+
 /**
  * pango_cairo_glyph_string_path
  * @cr: a Cairo context
@@ -416,30 +645,10 @@ pango_cairo_glyph_string_path (cairo_t          *cr,
 			       PangoFont        *font,
 			       PangoGlyphString *glyphs)
 {
-  PangoFontMap *fontmap;
-  PangoCairoRenderer *crenderer;
-  PangoRenderer *renderer;
-
   g_return_if_fail (cr != NULL);
-  g_return_if_fail (PANGO_IS_CAIRO_FONT (font));
   g_return_if_fail (glyphs != NULL);
 
-  fontmap = pango_font_get_font_map (font);
-  renderer = _pango_cairo_font_map_get_renderer (PANGO_CAIRO_FONT_MAP (fontmap));
-  crenderer = PANGO_CAIRO_RENDERER (renderer);
-
-  crenderer->cr = cr;
-  crenderer->do_path = TRUE;
-  cairo_get_current_point (cr, &crenderer->x_offset, &crenderer->y_offset);
-  
-  pango_renderer_draw_glyphs (renderer, font, glyphs, 0, 0);
-  
-  crenderer->cr = NULL;
-  crenderer->do_path = FALSE;
-  crenderer->x_offset = 0.;
-  crenderer->y_offset = 0.;
-  
-  cairo_set_font_face (cr, NULL);
+  _pango_cairo_do_glyph_string (cr, font, glyphs, TRUE);
 }
 
 /**
@@ -457,31 +666,10 @@ void
 pango_cairo_layout_line_path (cairo_t          *cr,
 			      PangoLayoutLine  *line)
 {
-  PangoContext *context;
-  PangoFontMap *fontmap;
-  PangoRenderer *renderer;
-  PangoCairoRenderer *crenderer;
-  
   g_return_if_fail (cr != NULL);
   g_return_if_fail (line != NULL);
 
-  context = pango_layout_get_context (line->layout);
-  fontmap = pango_context_get_font_map (context);
-  renderer = _pango_cairo_font_map_get_renderer (PANGO_CAIRO_FONT_MAP (fontmap));
-  crenderer = PANGO_CAIRO_RENDERER (renderer);
-
-  crenderer->cr = cr;
-  crenderer->do_path = TRUE;
-  cairo_get_current_point (cr, &crenderer->x_offset, &crenderer->y_offset);
-
-  pango_renderer_draw_layout_line (renderer, line, 0, 0);
-  
-  crenderer->cr = NULL;
-  crenderer->do_path = FALSE;
-  crenderer->x_offset = 0.;
-  crenderer->y_offset = 0.;
-  
-  cairo_set_font_face (cr, NULL);
+  _pango_cairo_do_layout_line (cr, line, TRUE);
 }
 
 /**
@@ -489,7 +677,7 @@ pango_cairo_layout_line_path (cairo_t          *cr,
  * @cr: a Cairo context
  * @layout: a Pango layout
  * 
- * Adds the text in a #PangoLayoutLine to the current path in the
+ * Adds the text in a #PangoLayout to the current path in the
  * specified cairo context.  The top-left corner of the #PangoLayout
  * will be at the current point of the cairo context.
  *
@@ -499,30 +687,37 @@ void
 pango_cairo_layout_path (cairo_t     *cr,
 			 PangoLayout *layout)
 {
-  PangoContext *context;
-  PangoFontMap *fontmap;
-  PangoRenderer *renderer;
-  PangoCairoRenderer *crenderer;
-
   g_return_if_fail (cr != NULL);
   g_return_if_fail (PANGO_IS_LAYOUT (layout));
 
-  context = pango_layout_get_context (layout);
-  fontmap = pango_context_get_font_map (context);
-  renderer = _pango_cairo_font_map_get_renderer (PANGO_CAIRO_FONT_MAP (fontmap));
-  crenderer = PANGO_CAIRO_RENDERER (renderer);
-  
-  crenderer->cr = cr;
-  crenderer->do_path = TRUE;
-  cairo_get_current_point (cr, &crenderer->x_offset, &crenderer->y_offset);
-
-  pango_renderer_draw_layout (renderer, layout, 0, 0);
-  
-  crenderer->cr = NULL;
-  crenderer->do_path = FALSE;
-  crenderer->x_offset = 0.;
-  crenderer->y_offset = 0.;
-  
-  cairo_set_font_face (cr, NULL);
+  _pango_cairo_do_layout (cr, layout, TRUE);
 }
 
+/**
+ * pango_cairo_error_underline_path:
+ * @cr: a Cairo context
+ * @x: The X coordinate of one corner of the rectangle
+ * @y: The Y coordinate of one corner of the rectangle
+ * @width: Non-negative width of the rectangle
+ * @height: Non-negative height of the rectangle
+ * 
+ * Add a squiggly line to the current path in the specified cairo context that
+ * approximately covers the given rectangle in the style of an underline used
+ * to indicate a spelling error.  (The width of the underline is rounded to an
+ * integer number of up/down segments and the resulting rectangle is centered
+ * in the original rectangle)
+ *
+ * Since: 1.14
+ **/
+void 
+pango_cairo_error_underline_path (cairo_t *cr,                                  
+			          double   x,
+			          double   y,
+			          double   width,
+			          double   height)
+{
+  g_return_if_fail (cr != NULL);
+  g_return_if_fail ((width >= 0) && (height >= 0));
+  
+  _pango_cairo_do_error_underline (cr, x, y, width, height, TRUE);                                 
+}
