@@ -26,7 +26,7 @@
 #include "pangofc-private.h"
 #include "pango-layout.h"
 #include "pango-modules.h"
-#include "pango-utils.h"
+#include "pango-impl-utils.h"
 
 #include <fontconfig/fcfreetype.h>
 
@@ -37,14 +37,33 @@ enum {
   PROP_PATTERN
 };
 
+typedef struct _GUnicharToGlyphCacheEntry GUnicharToGlyphCacheEntry;
+
+/* An entry in the fixed-size cache for the gunichar -> glyph mapping.
+ * The cache is indexed by the lower N bits of the gunichar (see
+ * GLYPH_CACHE_NUM_ENTRIES).  For scripts with few code points,
+ * this should provide pretty much instant lookups.
+ *
+ * The "ch" field is zero if that cache entry is invalid.
+ */
+struct _GUnicharToGlyphCacheEntry
+{
+  gunichar   ch;
+  PangoGlyph glyph;
+};
+
 typedef struct _PangoFcFontPrivate PangoFcFontPrivate;
 
 struct _PangoFcFontPrivate
 {
   PangoFcDecoder *decoder;
   gpointer context_key;
+  GUnicharToGlyphCacheEntry *char_to_glyph_cache;  
 };
 
+#define GLYPH_CACHE_NUM_ENTRIES 256 /* should be power of two */
+#define GLYPH_CACHE_MASK (GLYPH_CACHE_NUM_ENTRIES - 1)
+
 static gboolean pango_fc_font_real_has_char  (PangoFcFont *font,
 					      gunichar     wc);
 static guint    pango_fc_font_real_get_glyph (PangoFcFont *font,
@@ -64,14 +83,12 @@ static PangoFontMetrics *    pango_fc_font_get_metrics  (PangoFont        *font,
 							 PangoLanguage    *language);
 static PangoFontMap *        pango_fc_font_get_font_map (PangoFont        *font);
 static PangoFontDescription *pango_fc_font_describe     (PangoFont        *font);
+static PangoFontDescription *pango_fc_font_describe_absolute (PangoFont        *font);
 
 
 #define PANGO_FC_FONT_LOCK_FACE(font)	(PANGO_FC_FONT_GET_CLASS (font)->lock_face (font))
 #define PANGO_FC_FONT_UNLOCK_FACE(font)	(PANGO_FC_FONT_GET_CLASS (font)->unlock_face (font))
 
-
-
-
 G_DEFINE_TYPE (PangoFcFont, pango_fc_font, PANGO_TYPE_FONT)
 
 static void
@@ -87,6 +104,7 @@ pango_fc_font_class_init (PangoFcFontClass *class)
   object_class->finalize = pango_fc_font_finalize;
   object_class->set_property = pango_fc_font_set_property;
   font_class->describe = pango_fc_font_describe;
+  font_class->describe_absolute = pango_fc_font_describe_absolute;
   font_class->find_shaper = pango_fc_font_find_shaper;
   font_class->get_coverage = pango_fc_font_get_coverage;
   font_class->get_metrics = pango_fc_font_get_metrics;
@@ -134,6 +152,9 @@ pango_fc_font_finalize (GObject *object)
   if (priv->decoder)
     _pango_fc_font_set_decoder (fcfont, NULL);
 
+  if (priv->char_to_glyph_cache)
+    g_free (priv->char_to_glyph_cache);
+
   G_OBJECT_CLASS (pango_fc_font_parent_class)->finalize (object);
 }
 
@@ -208,6 +229,19 @@ pango_fc_font_describe (PangoFont *font)
   return pango_font_description_copy (fcfont->description);
 }
 
+static PangoFontDescription *
+pango_fc_font_describe_absolute (PangoFont *font)
+{
+  PangoFcFont *fcfont = (PangoFcFont *)font;
+  PangoFontDescription *desc = pango_font_description_copy (fcfont->description);
+  double size;
+
+  if (FcPatternGetDouble (fcfont->font_pattern, FC_PIXEL_SIZE, 0, &size) == FcResultMatch)
+    pango_font_description_set_absolute_size (desc, size * PANGO_SCALE);
+
+  return desc;
+}
+
 static PangoMap *
 pango_fc_get_shaper_map (PangoLanguage *language)
 {
@@ -229,9 +263,11 @@ pango_fc_font_find_shaper (PangoFont     *font,
 			   guint32        ch)
 {
   PangoMap *shaper_map = NULL;
+  PangoScript script;
 
   shaper_map = pango_fc_get_shaper_map (language);
-  return (PangoEngineShape *)pango_map_get_engine (shaper_map, ch);
+  script = pango_script_for_unichar (ch);
+  return (PangoEngineShape *)pango_map_get_engine (shaper_map, script);
 }
 
 static PangoCoverage *
@@ -252,28 +288,6 @@ pango_fc_font_get_coverage (PangoFont     *font,
 					  fcfont);
 }
 
-static void
-quantize_position (int *thickness,
-		   int *position)
-{
-  int thickness_pixels = (*thickness + PANGO_SCALE / 2) / PANGO_SCALE;
-  if (thickness_pixels == 0)
-    thickness_pixels = 1;
-  
-  if (thickness_pixels & 1)
-    {
-      int new_center = ((*position - *thickness / 2) & ~(PANGO_SCALE - 1)) + PANGO_SCALE / 2;
-      *position = new_center + (PANGO_SCALE * thickness_pixels) / 2;
-    }
-  else
-    {
-      int new_center = ((*position - *thickness / 2 + PANGO_SCALE / 2) & ~(PANGO_SCALE - 1));
-      *position = new_center + (PANGO_SCALE * thickness_pixels) / 2;
-    }
-
-  *thickness = thickness_pixels * PANGO_SCALE;
-}
-
 /* For Xft, it would be slightly more efficient to simply to
  * call Xft, and also more robust against changes in Xft.
  * But for now, we simply use the same code for all backends.
@@ -386,8 +400,10 @@ get_face_metrics (PangoFcFont      *fcfont,
    */
   if (fcfont->is_hinted)
     {
-      quantize_position (&metrics->underline_thickness, &metrics->underline_position);
-      quantize_position (&metrics->strikethrough_thickness, &metrics->strikethrough_position);
+      pango_quantize_line_geometry (&metrics->underline_thickness,
+				    &metrics->underline_position);
+      pango_quantize_line_geometry (&metrics->strikethrough_thickness,
+				    &metrics->strikethrough_position);
     }
   
   PANGO_FC_FONT_UNLOCK_FACE (fcfont);
@@ -426,13 +442,15 @@ pango_fc_font_create_metrics_for_context (PangoFcFont   *fcfont,
   PangoRectangle extents;
   PangoLanguage *language = pango_context_get_language (context);
   const char *sample_str = pango_language_get_sample_string (language);
+  PangoFontDescription *desc = pango_font_describe_with_absolute_size (fcfont);
   
   metrics = pango_font_metrics_new ();
   
   get_face_metrics (fcfont, metrics);
   
   layout = pango_layout_new (context);
-  pango_layout_set_font_description (layout, fcfont->description);
+  pango_layout_set_font_description (layout, desc);
+  pango_font_description_free (desc);
   
   pango_layout_set_text (layout, sample_str, -1);      
   pango_layout_get_extents (layout, NULL, &extents);
@@ -525,15 +543,35 @@ pango_fc_font_real_get_glyph (PangoFcFont *font,
   FT_Face face;
   FT_UInt index;
 
-  face = PANGO_FC_FONT_LOCK_FACE (font);
-  
-  index = FcFreeTypeCharIndex (face, wc);
-  if (index > (FT_UInt)face->num_glyphs)
-    index = 0;
+  guint idx;
+  GUnicharToGlyphCacheEntry *entry;
 
-  PANGO_FC_FONT_UNLOCK_FACE (font);
+  PangoFcFontPrivate *priv = font->priv;
 
-  return index;
+  if (G_UNLIKELY (priv->char_to_glyph_cache == NULL))
+    {
+      priv->char_to_glyph_cache = g_new0 (GUnicharToGlyphCacheEntry, GLYPH_CACHE_NUM_ENTRIES);
+      /* Make sure all cache entries are invalid initially */
+      priv->char_to_glyph_cache[0].ch = 1; /* char 1 cannot happen in bucket 0 */
+    }
+
+  idx = wc & GLYPH_CACHE_MASK;
+  entry = priv->char_to_glyph_cache + idx;
+
+  if (entry->ch != wc)
+    {
+      face = PANGO_FC_FONT_LOCK_FACE (font);
+      index = FcFreeTypeCharIndex (face, wc);
+      if (index > (FT_UInt)face->num_glyphs)
+        index = 0;
+
+      entry->ch = wc;
+      entry->glyph = index;
+      
+      PANGO_FC_FONT_UNLOCK_FACE (font);
+    }
+
+  return entry->glyph;
 }
 
 /**
